// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:wyatt_architecture/wyatt_architecture.dart';
import 'package:wyatt_i18n/src/core/utils/i18n_parser.dart';
import 'package:wyatt_i18n/wyatt_i18n.dart';
import 'package:wyatt_type_utils/wyatt_type_utils.dart' hide Option;

class I18nRepositoryImpl extends I18nRepository {
  I18nRepositoryImpl({
    required this.dataSource,
  }) : super();

  final I18nDataSource dataSource;
  I18n _i18n = const I18n.empty();

  Future<I18n> _parse(
    String content,
    Parser<String, Map<String, dynamic>> parser, {
    String? locale,
    bool strict = false,
  }) async {
    final parsed = parser.parse(content);

    String parsedLocale;

    /// Checks if the locale is present in the parsed data.
    /// If not, throws an exception.
    /// If yes, sets the locale to the parsed locale.
    if (parsed.containsKey('@@locale')) {
      if (strict) {
        /// Checks if the parsed locale is a string.
        if (parsed['@@locale'] is! String) {
          throw NoLocaleException();
        }

        /// Checks if the parsed locale is the same as the given locale.
        if (locale != null && parsed['@@locale'] as String != locale) {
          throw InvalidLocaleException(locale, parsed['@@locale'] as String);
        }
      }
      parsedLocale = parsed['@@locale'] as String;
    } else {
      if (strict) {
        /// Throws an exception if the locale is not present in the parsed data.
        throw NoLocaleException();
      } else {
        /// Sets the locale to the given locale.
        /// If the given locale is null, sets the locale to 'null'.
        /// This is done to prevent the locale from being null.
        /// It should never be null.
        parsedLocale = locale ?? 'null';
      }
    }

    return I18n(
      locale: parsedLocale,
      unparsedData: content,
      data: parsed,
    );
  }

  @override
  FutureOrResult<I18n> load({
    required String? locale,
    bool strict = false,
    Parser<String, Map<String, dynamic>>? parser,
  }) async =>
      await Result.tryCatchAsync<I18n, AppException, AppException>(
        () async {
          final content = await dataSource.load(locale: locale);

          return _i18n = await _parse(
            content,
            parser ?? dataSource.format.parser,
            locale: locale,
            strict: strict,
          );
        },
        (error) => error,
      );

  @override
  FutureOrResult<I18n> loadFrom(
    Uri uri, {
    Parser<String, Map<String, dynamic>>? parser,
  }) async =>
      await Result.tryCatchAsync<I18n, AppException, AppException>(
        () async {
          final content = await dataSource.loadFrom(uri);

          /// Strict is always true when loading from a uri. Because
          /// the locale is not given and can't be inferred.
          return _i18n = await _parse(
            content,
            parser ?? dataSource.format.parser,
            strict: true,
          );
        },
        (error) => error,
      );

  @override
  Result<String, AppException> get(
    String key, [
    Map<String, dynamic> arguments = const {},
  ]) {
    final I18nParser parser = I18nParser(i18n: _i18n, arguments: arguments);

    if (_i18n.containsKey(key)) {
      return Result.tryCatch<String, AppException, AppException>(
        () => parser.parse(_i18n[key] as String, key: key),
        (error) => error,
      );
    } else {
      throw KeyNotFoundException(key, arguments);
    }
  }
}
